/*
 * NBCTK
 * Copyright (c) 2007-2011, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

/*
 * The memory managers specific to matrices/vectors, which is based on
 * the one described in the K&R book (Chap. 8.7).  Also, following the
 * IEEE standard, we assume that a zero value (0.0) in a matrix/vector
 * is represented by all zero bits.
 *
 * [Note]
 * When running with OpenMP (i.e. _OPENMP is defined), a matrix/vector
 * that is created by a thread (with alloc_matrix/vector()) should be
 * released by the same thread (with release_matrix/vector()).
 */

#ifdef _OPENMP
static M_POOL **mroot0;
static V_POOL **vroot0;
#else
static M_POOL *mroot0;
static V_POOL *vroot0;
#endif

#ifdef _OPENMP
/* `th' should be declared as a local variable */
#  define mroot (mroot0[th])
#  define vroot (vroot0[th])
#else
#  define mroot mroot0
#  define vroot vroot0
#endif

M_POOL *alloc_matrix_pool(int size)
{
  M_POOL *pool;
  M_HEADER *h;

  pool = (M_POOL *)xmalloc(sizeof(M_POOL));

  h = (M_HEADER *)xmalloc(sizeof(M_HEADER)*size);
  memset(h,0, sizeof(M_HEADER)*size);
  h->info.dim0 = 0;
  h->info.dim1 = 0;
  h->info.next = h;

  pool->pool = h;
  pool->free = h;
  pool->last = h + 1;
  pool->size = size;

  return pool;
}

V_POOL *alloc_vector_pool(int size)
{
  V_POOL *pool;
  V_HEADER *h;

  pool = (V_POOL *)xmalloc(sizeof(V_POOL));

  h = (V_HEADER *)xmalloc(sizeof(V_HEADER)*size);
  memset(h,0, sizeof(V_HEADER)*size);
  h->info.dim  = 0;
  h->info.next = h;

  pool->pool = h;
  pool->free = h;
  pool->last = h + 1;
  pool->size = size;

  return pool;
}

#ifdef _OPENMP
void initialize_matrix_pool(void)
{
  int nth = omp_get_max_threads(), th;

  mroot0 = (M_POOL **)xmalloc(sizeof(M_POOL *)*nth);
  for (th=0; th<nth; th++) {
    mroot0[th] = alloc_matrix_pool(M_POOL_SIZE);
    mroot0[th]->next = NULL;
  }
}
#else
void initialize_matrix_pool(void)
{
  mroot0 = alloc_matrix_pool(M_POOL_SIZE);
  mroot0->next = NULL;
}
#endif

#ifdef _OPENMP
void initialize_vector_pool(void)
{
  int nth = omp_get_max_threads(), th;

  vroot0 = (V_POOL **)xmalloc(sizeof(V_POOL *)*nth);
  for (th=0; th<nth; th++) {
    vroot0[th] = alloc_vector_pool(V_POOL_SIZE);
    vroot0[th]->next = NULL;
  }
}
#else
void initialize_vector_pool(void)
{
  vroot0 = alloc_vector_pool(V_POOL_SIZE);
  vroot0->next = NULL;
}
#endif

#ifdef _OPENMP
void release_matrix_pool(void)
{
  int nth = omp_get_max_threads(), th;
  int n0,nunits;
  M_POOL *pool,*pool0;
  M_HEADER *mp;
  FLOAT_T **p;

  for (th=0; th<nth; th++) {
    for (pool = mroot0[th]; pool != NULL; pool = pool->next) {
      mp = pool->pool;
      while (mp < pool->last) {
        n0 = mp->info.dim0;

        if (n0 > 0) {
          p = (FLOAT_T **)(mp + 1);
          xfree(p[0]);
          nunits =
            (n0 * sizeof(FLOAT_T *) + sizeof(M_HEADER) - 1)/sizeof(M_HEADER) + 1;
        }
        else nunits = 1;

        mp += nunits;
      }
      xfree(pool->pool);
      pool->size = 0;
      pool->free = NULL;
      pool->last = NULL;
    }

    pool = mroot0[th];
    while (pool != NULL) {
      pool0 = pool;
      pool = pool->next;
      xfree(pool0);
    }
    mroot0[th] = NULL;
  }

  xfree(mroot0);
}
#else
void release_matrix_pool(void)
{
  int n0,nunits;
  M_POOL *pool,*pool0;
  M_HEADER *mp;
  FLOAT_T **p;

  for (pool = mroot0; pool != NULL; pool = pool->next) {
    mp = pool->pool;
    while (mp < pool->last) {
      n0 = mp->info.dim0;

      if (n0 > 0) {
        p = (FLOAT_T **)(mp + 1);
        xfree(p[0]);
        nunits =
          (n0 * sizeof(FLOAT_T *) + sizeof(M_HEADER) - 1)/sizeof(M_HEADER) + 1;
      }
      else nunits = 1;

      mp += nunits;
    }
    xfree(pool->pool);
    pool->size = 0;
    pool->free = NULL;
    pool->last = NULL;
  }

  pool = mroot0;
  while (pool != NULL) {
    pool0 = pool;
    pool = pool->next;
    xfree(pool0);
  }
  mroot0 = NULL;
}
#endif

#ifdef _OPENMP
void release_vector_pool(void)
{
  int nth = omp_get_max_threads(), th;
  V_POOL *pool,*pool0;

  for (th=0; th<nth; th++) {
    for (pool = vroot0[th]; pool != NULL; pool = pool->next) {
      xfree(pool->pool);
      pool->size = 0;
      pool->free = NULL;
      pool->last = NULL;
    }

    pool = vroot0[th];
    while (pool != NULL) {
      pool0 = pool;
      pool = pool->next;
      xfree(pool0);
    }
    vroot0[th] = NULL;
  }

  xfree(vroot0);
}
#else
void release_vector_pool(void)
{
  V_POOL *pool,*pool0;

  for (pool = vroot0; pool != NULL; pool = pool->next) {
    xfree(pool->pool);
    pool->size = 0;
    pool->free = NULL;
    pool->last = NULL;
  }

  pool = vroot0;
  while (pool != NULL) {
    pool0 = pool;
    pool = pool->next;
    xfree(pool0);
  }
  vroot0 = NULL;
}
#endif

MATRIX alloc_matrix(int n0, int n1)
{
#ifdef _OPENMP
  int th = omp_get_thread_num();
#endif
  M_POOL *pool;
  M_HEADER *mp,*mp_prev;
  int i,nunits,size;
  FLOAT_T **p,*q;

#ifdef CHECK_DIM
  if (n0 <= 0 || n1 <= 0) DIE_DIM;
#endif

  nunits =
    (n0 * sizeof(FLOAT_T *) + sizeof(M_HEADER) - 1)/sizeof(M_HEADER) + 1;

  /* Firstly, search for a freed area... */

  for (pool = mroot; pool != NULL; pool = pool->next) {
    mp_prev = pool->free;
    mp = mp_prev->info.next;

    while (1) {
      if (mp->info.dim0 == n0 && mp->info.dim1 == n1) {
        mp_prev->info.next = mp->info.next;
        pool->free = mp_prev;

        return (MATRIX)(mp + 1);
      }

      /* No freed area found in the current pool */
      if (mp == pool->free) break;

      mp_prev = mp;
      mp = mp->info.next;
    }
  }

  /* Secondly, allocate an area in an existing pool... */

  for (pool = mroot; pool != NULL; pool = pool->next) {
    if ((pool->last + nunits - pool->pool) < pool->size) {
      pool->last->info.dim0 = n0;
      pool->last->info.dim1 = n1;

      p = (FLOAT_T **)(pool->last + 1);
      q = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*n0*n1);
      for (i=0; i<n0; i++)
        p[i] = q + i * n1;

      pool->last += nunits;

      return (MATRIX)p;
    }
  }

  /* Thirdly, we create a new pool... */

  size = ((mroot->size > nunits) ? mroot->size : nunits) * 2;
  pool = alloc_matrix_pool(size);
  pool->next = mroot;  /* larger pools come nearer to the root */
  mroot = pool;

  /* and then, allocate an area from the new pool... */

  pool->last->info.dim0 = n0;
  pool->last->info.dim1 = n1;

  p = (FLOAT_T **)(pool->last + 1);
  q = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*n0*n1);
  for (i=0; i<n0; i++)
    p[i] = q + i * n1;

  pool->last += nunits;

  return (MATRIX)p;
}

VECTOR alloc_vector(int n)
{
#ifdef _OPENMP
  int th = omp_get_thread_num();
#endif
  V_POOL *pool;
  V_HEADER *vp,*vp_prev;
  int nunits,size;
  VECTOR p;

#ifdef CHECK_DIM
  if (n <= 0) DIE_DIM;
#endif

  nunits = (n * sizeof(FLOAT_T) + sizeof(V_HEADER) - 1)/sizeof(V_HEADER) + 1;

  /* Firstly, search for a freed area... */

  for (pool = vroot; pool != NULL; pool = pool->next) {
    vp_prev = pool->free;
    vp = vp_prev->info.next;

    while (1) {
      if (vp->info.dim == n) {
        vp_prev->info.next = vp->info.next;
        pool->free = vp_prev;

        return (VECTOR)(vp + 1);
      }

      /* No freed area found in the current pool */
      if (vp == pool->free) break;

      vp_prev = vp;
      vp = vp->info.next;
    }
  }

  /* Secondly, allocate an area in an existing pool... */

  for (pool = vroot; pool != NULL; pool = pool->next) {
    if ((pool->last + nunits - pool->pool) < pool->size) {
      pool->last->info.dim = n;
      p = (VECTOR)(pool->last + 1);
      pool->last += nunits;
      return p;
    }
  }

  /* Thirdly, we create a new pool... */

  size = ((vroot->size > nunits) ? vroot->size : nunits) * 2;
  pool = alloc_vector_pool(size);
  pool->next = vroot;  /* larger pools come nearer to the root */
  vroot = pool;

  /* and then, allocate an area from the new pool... */

  pool->last->info.dim = n;
  p = (VECTOR)(pool->last + 1);
  pool->last += nunits;

  return (VECTOR)p;
}

void release_matrix0(MATRIX m)
{
#ifdef _OPENMP
  int th = omp_get_thread_num();
#endif
  M_POOL *pool;
  M_HEADER *mp,*b;

  if (m == NULL) return;

  b = (M_HEADER *)m - 1;

  for (pool = mroot; pool != NULL; pool = pool->next) {
    if (b >= pool->pool && b < pool->last) {
      mp = pool->free;

      while (1) {
        if (b == mp) die("Internal error: double-freed a matrix (%p)\n",b);
        if ((b > mp && b < mp->info.next) ||
            (mp >= mp->info.next && (b > mp || b < mp->info.next))) break;
        mp = mp->info.next;
      }

      b->info.next = mp->info.next;
      mp->info.next = b;

      pool->free = b;

      return;
    }
  }

  die("Internal error: Illegal address of a matrix to be freed\n");
}

void release_vector0(VECTOR v)
{
#ifdef _OPENMP
  int th = omp_get_thread_num();
#endif
  V_POOL *pool;
  V_HEADER *vp,*b;

  if (v == NULL) return;

  b = (V_HEADER *)v - 1;

  for (pool = vroot; pool != NULL; pool = pool->next) {
    if (b >= pool->pool && b < pool->last) {
      vp = pool->free;

      while (1) {
        if (b == vp) die("Internal error: double-freed a vector (%p)\n",b);
        if ((b > vp && b < vp->info.next) ||
            (vp >= vp->info.next && (b > vp || b < vp->info.next))) break;
        vp = vp->info.next;
      }

      b->info.next = vp->info.next;
      vp->info.next = b;

      pool->free = b;

      return;
    }
  }

  die("Internal error: Illegal address of a vector to be freed\n");
}
